''
''
''     QuadRotor - Full Stability
'' -- Jason Dorie --               
''               
''

{
  TODO:
        Use integrated gyro value for P, current gyro measurement for D
        Use accelerometer to bias integrated gyro result
                Store last 128 results in HUB RAM, keep a running sum, SHR 7 for current value

  DONE:
        Temperature drift compensation
        Convert ITG-3200 code to PASM
        Add ADXL345 readings to ITG code
        Integration of gyro values
}


CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  OUT_FL = 0                ' Front Left ESC                       
  OUT_FR = 1                ' Front Right ESC
  OUT_BL = 2                ' Back Left ESC                        
  OUT_BR = 3                ' Back Right ESC



OBJ
  Gyro  : "ITG-3200-pasm.spin"                          ' 1 cog
  RC    : "RC_Receiver_4.spin"                          ' 1 cog                       
  ESC   : "Servo8Fast.spin"                             ' 1 cog
  'Dbg   : "FullDuplexSerial.spin"
  'XBee  : "FullDuplexSerial.spin"

  RollPID   : "IntPID.spin"
  PitchPID  : "IntPID.spin"
  YawPID    : "IntPID.spin"

VAR
  long Output[4]                                        ' Temp output flight controls array, copied into 'servo' when finished                        
  long counter
  long Thro, Aile, Elev, Rudd
  long iAile, iElev, iRudd
  long Roll, Pitch, Yaw
  long DesiredRoll, DesiredPitch, DesiredYaw
  long RollOut, PitchOut, YawOut
  long ThroMix
  long LastTime, CurTime

  long Gx, Gy, Gz                                       ' Gyro readings (instant)
  long iGx, iGy, iGz                                    ' Gyro readings (integrated)

  WORD POut
  WORD DOut

PUB Main
 
  'Dbg.Start( 31, 30, 0, 115200 )
  'XBee.Start( 24, 25, 0, 115200 )
  Gyro.Start( 16, 17 )


  ' Initialize servo values to all neutral (zero for throttles)
  ' Delay values of 1500 = center, 1000 = min, 2000 = max
  Output[0] := 1100    
  Output[1] := 1100              ' motors off    
  Output[2] := 1100    
  Output[3] := 1100

  'Start the servo driver cog
  ESC.AddPin(8)
  ESC.AddPin(9)
  ESC.AddPin(10)
  ESC.AddPin(11)  
  ESC.Set(  8, Output[0] )
  ESC.Set(  9, Output[1] )
  ESC.Set( 10, Output[2] )
  ESC.Set( 11, Output[3] )
  ESC.Start

  'Initialize the RC input object
  RC.setpins( %1111 )
  RC.start  


  '-----Original settings -------  
  'RollPID.Init(  60, 0, 10 )
  'PitchPID.Init( 60, 0, 10 )                       
  'YawPID.Init(   60, 0, 10 )                    
  '-----Original settings -------  


  '-----Good settings-----
  'RollPID.Init(  12, 0, 12 )
  'PitchPID.Init( 12, 0, 12 )                       
  'YawPID.Init(   30, 0, 15 )                    
  '-----Good settings-----


  RollPID.Init(  12, 0, 12 )
  PitchPID.Init( 12, 0, 12 )                       
  YawPID.Init(   40, 0, 25 )                    

  'Should probably wait for the gyro to finish startup here

  'repeat until( Gyro.IsReady == 1 )

  FlightLoop



PUB FlightLoop

  counter := 0
  LastTime := cnt

  Roll := Pitch := Yaw := 0
  DesiredRoll := DesiredPitch := DesiredYaw := 0

  repeat
    Thro := RC.getrc(0)
    Aile := RC.getrc(1)
    Elev := RC.getrc(2)
    Rudd := RC.getrc(3)

    iAile += Aile
    iElev += Elev
    iRudd += Rudd

    Gy := Gyro.GetRX           ' Gyro is mounted with a 90 deg rotation
    Gx := Gyro.GetRY
    Gz := Gyro.GetRZ

    iGx += Gx
    iGy += Gy
    iGz += Gz

    'Zero all targets when throttle is off - makes for more stable liftoff
    if( Thro < -380 )
      iGx := iAile := 0
      iGy := iElev := 0
      iGz := iRudd := 0

    DesiredRoll := iAile << 1
    DesiredPitch := iElev << 1
    DesiredYaw := iRudd << 1


    'Dbg.Dec( iGz )
    'Dbg.tx( 32 )
    'Dbg.Dec( Gz )
    'Dbg.tx( 13 )

    {
    counter++
    if( counter == 100 )
      dbg.dec( Thro )
      dbg.tx( 32 )
      dbg.dec( DesiredRoll )
      dbg.tx( 32 )
      dbg.dec( DesiredPitch )
      dbg.tx(13)
      counter := 0
    '}  


    RollOut := RollPID.Calculate( DesiredRoll , iGx )    
    PitchOut := PitchPID.Calculate( DesiredPitch , iGy )    
    YawOut := YawPID.Calculate( DesiredYaw , iGz )    


    {
    counter++
    if( counter == 3 )
      POut := RollPID.GetPOut
      DOut := RollPID.GetDOut
      
      XBee.tx( $7f )
      XBee.tx( $7f )
      XBee.tx( DesiredRoll >> 12 )
      XBee.tx( (DesiredRoll >> 4) & $ff )
      XBee.tx( iGx >> 12 )
      XBee.tx( (iGx >> 4) & $ff )
      XBee.tx( POut >> 8 )
      XBee.tx( POut & $ff )
      XBee.tx( DOut >> 8 )
      XBee.tx( DOut & $ff )
      counter := 0
    '}   


    ThroMix := (Thro + 400) ~> 2                        ' Approx 0 - 256
    ThroMix <#= 64                                      ' Above 1/4 throttle, clamp it to 64
    ThroMix #>= 0

    'add 1500 to all Output values to make them 'servo friendly' again
    Thro += 1500


    ' + configuration
    'Output[OUT_F] := Thro + ((-PitchOut + YawOut) * ThroMix) ~> 6                          
    'Output[OUT_R] := Thro + (( RollOut  - YawOut) * ThroMix) ~> 6  
    'Output[OUT_B] := Thro + (( PitchOut + YawOut) * ThroMix) ~> 6
    'Output[OUT_L] := Thro + ((-RollOut  - YawOut) * ThroMix) ~> 6

    ' X configuration
    Output[OUT_FL] := Thro + ((-PitchOut - RollOut - YawOut) * ThroMix) ~> 6                          
    Output[OUT_FR] := Thro + ((-PitchOut + RollOut + YawOut) * ThroMix) ~> 6  
    Output[OUT_BL] := Thro + (( PitchOut - RollOut + YawOut) * ThroMix) ~> 6
    Output[OUT_BR] := Thro + (( PitchOut + RollOut - YawOut) * ThroMix) ~> 6


    {
    counter++
    if( counter == 50 )
      dbg.dec( Output[OUT_FL] )
      dbg.tx( 32 )
      dbg.dec( Output[OUT_FR] )
      dbg.tx( 32 )
      dbg.dec( Output[OUT_BL] )
      dbg.tx( 32 )
      dbg.dec( Output[OUT_BR] )
      dbg.tx(13)
      counter := 0
    '}  
     

    Output[0] #>= 1000
    Output[0] <#= 2000
    Output[1] #>= 1000
    Output[1] <#= 2000
    Output[2] #>= 1000
    Output[2] <#= 2000
    Output[3] #>= 1000
    Output[3] <#= 2000
     
    'Copy new Ouput array into servo values
    ESC.Set(  8, Output[0] )
    ESC.Set(  9, Output[1] )
    ESC.Set( 10, Output[2] )
    ESC.Set( 11, Output[3] )

    waitcnt( constant(80_000_000 / 200) + LastTime )
    LastTime += constant(80_000_000 / 200)